#include "./LCPscan.h"

#include <stxxl/bits/containers/sorter.h>
#include <stxxl/bits/containers/queue.h>
#include <stxxl/bits/mng/buf_istream.h>
#include <stxxl/bits/mng/buf_ostream.h>

#include <cstdio>
#include <cstdlib>

#include <list>
#include <string>
#include <numeric>
#include <algorithm>

#include "./stream.h"
#include "./uint40.h"
#include "./utils.h"
#include "./tuples.h"

// Main consts.
const long kDiskBlockSize = (1 << 20);  // disk block size
const long kSBlockCnt = 4;  // number of superblocks
const long double kSBlockFractions[kSBlockCnt] = {0.31L, 0.27L, 0.23L, 0.19L};
const long kPrintStep = (1L << 25) - 1;
const long double kMiB = 1024.L * 1024;
const long kSmallInstance = (1 << 10);

// Basic statistics.
long max_lcp;
long n_irr;
long lcp_sum;
long irr_lcp_sum;

// Return peak STXXL disk space allocation (in bytes).
long max_alloc() {
  return stxxl::block_manager::get_instance()->max_allocated();
}

typedef stxxl::sorter<Triple, CmpTriple, kDiskBlockSize> triple_sorter_type;
typedef stxxl::sorter<Pair, CmpPair, kDiskBlockSize> pair_sorter_type;

//==============================================================================
// Compute a set crit_pos containing all positions i such that PLCP[i] = 0.
//==============================================================================
void compute_crit_pos(std::string text_fname, std::string sa_fname,
    std::list<long> &crit_pos) {
  fprintf(stderr, "\rCounting characters:");
  long double start = utils::wclock();

  // Compute symbol frequencies.
  long char_count[257] = {0L}, text_length = 0L;
  stream_reader<unsigned char> *sr =
      new stream_reader<unsigned char>(text_fname);
  for (; !(sr->empty()); ++text_length) {
    ++char_count[size_t(sr->read()) + 1];

    if (!(text_length & kPrintStep)) {
      long double elapsed = utils::wclock() - start;
      fprintf(stderr, "\rCounting characters: time = %.1LFs, io = %.2LfMiB/s",
        elapsed, (text_length / kMiB) / elapsed);
    }
  }
  delete sr;

  int sigma = 0;  // alphabet size
  for (int i = 1; i <= 256; ++i)
    if (char_count[i]) ++sigma;

  std::partial_sum(char_count, char_count + 257, char_count);

  long double elapsed = utils::wclock() - start;
  fprintf(stderr, "\rCounting characters: "
      "time = %.2Lfs "
      "(+%.3Lfs/MiB), "
      "io = %.2LfMiB/s\n",
      elapsed,
      elapsed / (text_length / kMiB),
      (text_length / kMiB) / elapsed);
  fprintf(stderr, "Alphabet size = %d\n\n", sigma);

  std::FILE *fsa = utils::file_open(sa_fname, "r");
  int n_unique = std::unique(char_count, char_count + 256) - char_count;
  if (char_count[n_unique - 1] == text_length) --n_unique;

  // LCP[i] = 0 iff char_count[0..n_unique) contains i.
  for (int i = 0; i < n_unique; ++i)
    crit_pos.push_back(
        utils::read_object_at_offset<uint40>(fsa, char_count[i]).ll() + 1);
  std::fclose(fsa);

  crit_pos.push_back(text_length);
  crit_pos.sort();
}

//==============================================================================
// Presort triples (SA[i], SA[i - 1], i) such that SA[i] belongs to interval
// [beg, end). The sorting key is the first element of the triple.
//
// After sorting:
//   * second components form Phi[beg, end),
//   * third components form ISA[beg, end).
//
// Output remains accessible through the triple_sorter passed as an argument.
// Returns: minimal budget necessary to merge runs in one pass.
//==============================================================================
long sort_triples(std::string sa_fname, long text_length, long beg, long end,
    triple_sorter_type &triple_sorter) {
  fprintf(stderr, "\r  Sorting triples:");
  long double start = utils::wclock();

  // Scan SA and for every value in range create a triple.
  stream_reader<uint40> SA(sa_fname);
  for (long i = 0L, prev = text_length; i < text_length; ++i) {
    long sai = SA.read();
    if (beg <= sai && sai < end)
      triple_sorter.push(Triple(sai, prev, i));
    prev = sai;

    // Print progress.
    if (!(i & kPrintStep)) {
      long double elapsed = utils::wclock() - start;
      long in_volume = sizeof(uint40) * i;
      long out_volume = sizeof(Triple) * triple_sorter.size();
      long io_volume = in_volume + out_volume;
      long double io_speed = (io_volume / kMiB) / elapsed;
      fprintf(stderr, "\r  Sorting triples: %.1Lf%%, time = %.1Lfs, "
          "io = %.2LfMiB/s", (100.L * i) / text_length, elapsed, io_speed);
    }
  }

  long double elapsed = utils::wclock() - start;
  long io_volume = sizeof(Triple) * (end - beg) + 5L * text_length;
  long double io_speed = (io_volume / kMiB) / elapsed;
  fprintf(stderr, "\r  Sorting triples: time = %.1Lfs (+%.3Lfs/MiB), "
      "io = %.2LfMiB/s, max-alloc = %.2LfB/B\n", elapsed, elapsed /
      (text_length / kMiB), io_speed, 1.L * max_alloc() / text_length);

  long min_merge_budget = triple_sorter.min_merge_budget();
  triple_sorter.sort(min_merge_budget);
  return min_merge_budget;
}

//==============================================================================
// Presort lexicographically pairs (Phi[i], i) for all i such that i belongs to
// superblock [sb_beg, sb_end) and PLCP[i] is irreducible.
//
// Pairs (Phi[i], i) and (Phi[j], j) are handled by separate sorters if i and j
// belong to different blocks.
//
// INPUT:
//   Sequence of triples (i, Phi[i], ISA[i]) passed via triple_sorter.
//
// OUTPUT:
//   As described above, returned via pair_sorters.
//   The function also stores ISA[sb_beg..sb_end) to disk.
//==============================================================================
void sort_pairs(long sb_beg, long sb_end, long max_block_size, long text_length,
    triple_sorter_type *triple_sorter, pair_sorter_type** &pair_sorters,
    stxxl::BIDArray<kDiskBlockSize>* &isa_bids, long budget,
    std::list<long> &crit_pos) {
  long sb_size = sb_end - sb_beg;

  fprintf(stderr, "\r  Sorting pairs:");
  long double start = utils::wclock();

  // Allocate disk blocks to keep ISA.
  typedef stxxl::typed_block<kDiskBlockSize, uint40> bt;
  typedef stxxl::buf_ostream<bt,
    stxxl::BIDArray<kDiskBlockSize>::iterator> isa_ostream_type;
  long isa_disk_blocks = (sb_size + bt::size - 1) / bt::size;
  isa_bids = new stxxl::BIDArray<kDiskBlockSize>(isa_disk_blocks);
  stxxl::block_manager *bm = stxxl::block_manager::get_instance();
  bm->new_blocks(stxxl::striping(), isa_bids->begin(), isa_bids->end());

  // Create ISA writer.
  isa_ostream_type isa_ostream(isa_bids->begin(), 2);

  CmpPair cmp;
  long prev = text_length, cnt = 0L, irr = 0L;
  for (long b_beg = sb_beg, t = 0L; b_beg < sb_end; b_beg += max_block_size) {
    long b_end = std::min(sb_end, b_beg + max_block_size);

    pair_sorters[t] = new pair_sorter_type(cmp, budget, 0L);
    for (long i = b_beg; i < b_end; ++i, ++(*triple_sorter), ++cnt) {
      Triple p = *(*triple_sorter);
      // p.first = i,
      // p.second = Phi[i],
      // p.third = ISA[i].

      // Check if PLCP[i] is irreducible.
      if (crit_pos.front() == i || p.second.ll() != prev + 1) {
        ++irr;
        pair_sorters[t]->push(Pair(p.second, p.first));
        if (crit_pos.front() == i) crit_pos.pop_front();
      }
      prev = p.second;

      // Store ISA[i] on disk, we will need it to permute PLCP to LCP.
      isa_ostream << p.third;

      if (!(cnt & kPrintStep)) {
        long double elapsed = utils::wclock() - start;
        long in_volume = 15L * cnt;
        long out_volume = 5L * cnt + 10L * irr;
        long io_volume = in_volume + out_volume;
        long double io_speed = (io_volume / kMiB) / elapsed;
        fprintf(stderr, "\r  Sorting pairs: %.1Lf%%, time = %.1Lfs, "
            "io = %.2LfMiB/s", (100.L * cnt) / sb_size, elapsed, io_speed);
      }
    }

    pair_sorters[t++]->finish();
  }

  triple_sorter->finish_clear();
  delete triple_sorter;
  isa_ostream.fill(0);
  n_irr += irr;

  long double elapsed = utils::wclock() - start;
  long io_volume = 20L * cnt + 10L * irr;
  long double io_speed = (io_volume / kMiB) / elapsed;
  fprintf(stderr, "\r  Sorting pairs: time = %.1Lfs (+%.3Lfs/MiB), "
      "io = %.2LfMiB/s, max-alloc = %.2LfB/B\n", elapsed, elapsed /
      (text_length / kMiB), io_speed, 1.L * max_alloc() / text_length);
}

//==============================================================================
// Return length of lcp between suffixes starting at positions i and j.
//==============================================================================
long naive_lcp(long i, long j, long lcp, std::FILE *f_text, long text_length) {
  static const long bufsize = (1 << 20);
  unsigned char *b1 = new unsigned char[bufsize];
  unsigned char *b2 = new unsigned char[bufsize];

  while (true) {
    long toread = std::min(bufsize, text_length - std::max(i, j) - lcp);
    if (!toread) break;

    utils::read_objects_at_offset<unsigned char>(f_text, i + lcp, toread, b1);
    utils::read_objects_at_offset<unsigned char>(f_text, j + lcp, toread, b2);

    long lcp2 = 0L;
    while (lcp2 < toread && b1[lcp2] == b2[lcp2])
      ++lcp2;
    lcp += lcp2;

    if (lcp2 < toread)
      break;
  }
  delete[] b1;
  delete[] b2;

  return lcp;
}

//==============================================================================
// Process text block [b..e).
//
// INPUT
//   Pairs (i, Phi[i]) where PLCP[i] is irreducible via presorted
//   phi_pair_sorter.
//
// OUTPUT
//   The collection of pairs (i, PLCP[i]) which are fed to the sorter and
//   returned.
//==============================================================================
pair_sorter_type *process_block(long s, long e, std::string text_fname,
    long text_length, pair_sorter_type *pair_sorter) {
  fprintf(stderr, "\r  Processing block [%ld..%ld)", s, e);
  long double start = utils::wclock();

  // Create the output sorter.
  CmpPair cmp;
  pair_sorter_type *output = new pair_sorter_type(cmp, kDiskBlockSize * 40L, 0);

  // Initialize merging.
  // NOTE: (e - s) was chosen so that the space required by this merger, block
  // and space needed for sorting output pairs fit in the ram limit.
  pair_sorter->sort(pair_sorter->min_merge_budget());

  if (pair_sorter->empty()) {
    output->finish();
    return output;
  }

  // Read text block.
  long ee = std::min(text_length, e + kDiskBlockSize);  // add one disk block
  unsigned char *b = new unsigned char[ee - s];         // to detect long lcps
  utils::read_objects_at_offset(text_fname, s, ee - s, b);

  long j_prev = 0L, lcp_prev = 0L;
  long last = 0L, cnt = 0L;  // to print debug

  stxxl::queue<Triple> long_lcps;  // for lcps crossing block boundary
  stream_reader<unsigned char> t(text_fname);  // text reader

  for (; !pair_sorter->empty(); ++(*pair_sorter), ++cnt) {
    long i = (*pair_sorter)->second;  // s <= i < e
    long j = (*pair_sorter)->first;  // Phi[i]
    long lcp = 0L;

    lcp = std::max(lcp, lcp_prev + j_prev - j);  // PLCP property
    while (i + lcp < ee && j + lcp < text_length &&
        t[j + lcp] == b[i + lcp - s]) ++lcp;  // extend the match

    if (i + lcp >= ee && j + lcp < text_length)
      long_lcps.push(Triple(i, j, lcp));  // lcp crosses block boundary
    else
      output->push(Pair(i, lcp));  // PLCP[i] = lcp

    j_prev = j;
    lcp_prev = lcp;

    if (j - last > (kPrintStep << 3)) {
      last = j;
      long double elapsed = utils::wclock() - start;
      long io_vol = 10L * (cnt + output->size()) + j;
      long double io_speed = (io_vol / kMiB) / elapsed;
      fprintf(stderr, "\r  Processing block [%ld..%ld): %.1Lf%%, "
          "time = %.1Lfs, io = %.1LfMiB/s", s, e,
          (100.L * j) / text_length, elapsed, io_speed);
    }
  }

  delete[] b;
  pair_sorter->finish_clear();
  delete pair_sorter;

  long double elapsed = utils::wclock() - start;
  long io_vol = 10L * (cnt + output->size()) + text_length;
  long double io_speed = (io_vol / kMiB) / elapsed;
  fprintf(stderr, "\r  Processing block [%ld..%ld): "
      "time = %.1Lfs (+%.3Lfs/MiB), io = %.1LfMiB/s, "
      "max-alloc = %.2LfB/B\n", s, e, elapsed, elapsed / (text_length / kMiB),
      io_speed, 1.L * max_alloc() / text_length);

  // Compute long lcps (crossing the block boundary).
  if (!long_lcps.empty()) {
    start = utils::wclock();
    std::FILE *f_text = utils::file_open(text_fname, "r");
    for (; !long_lcps.empty(); long_lcps.pop()) {
      long i = long_lcps.front().first;
      long j = long_lcps.front().second;
      long lcp = long_lcps.front().third;

      lcp = naive_lcp(i, j, lcp, f_text, text_length);
      output->push(Pair(i, lcp));  // PLCP[i] = lcp
    }
    std::fclose(f_text);
    elapsed = utils::wclock() - start;
    fprintf(stderr, "  Computing long lcps: %.1Lfs (+%.3Lfs/MiB)\n",
          elapsed, elapsed / (text_length / kMiB));
  }

  output->finish();
  return output;
}

//==============================================================================
// Permute PLCP[sb_beg..sb_end) into lex order and save to disk.
//==============================================================================
void plcp_to_lcp(long sb_beg, long sb_end, long sb_id, long max_block_size,
    pair_sorter_type** &lcp_sorters, stxxl::BIDArray<kDiskBlockSize>* &isa_bids,
    stxxl::BIDArray<kDiskBlockSize>** &lcp_bids, long ram_use,
    long text_length) {
  long sb_size = sb_end - sb_beg;
  long blocks_in_sb = (sb_size + max_block_size - 1) / max_block_size;

  long budget = ram_use;
  for (long i = 0L; i < blocks_in_sb; ++i)
    budget = std::min(budget, ram_use - lcp_sorters[i]->min_merge_budget());
  fprintf(stderr, "  Budget for permuting PLCP into LCP = %.2LfMiB\n",
      budget / kMiB);

  fprintf(stderr, "\r  Permuting PLCP into LCP:");
  long double start = utils::wclock();

  CmpPair cmp;
  pair_sorter_type *lex_sorter =
    new pair_sorter_type(cmp, (20L * budget) / 31L, 0);

  typedef stxxl::typed_block<kDiskBlockSize, uint40> isa_block_type;
  typedef stxxl::buf_istream<isa_block_type,
    stxxl::BIDArray<kDiskBlockSize>::iterator> isa_istream_type;
  isa_istream_type isa(isa_bids->begin(), isa_bids->end(), 2);

  long prev = 0L, b_id = 0L, cnt = 0L, irr = 0L;
  for (long b_beg = sb_beg; b_beg < sb_end; b_beg += max_block_size) {
    long b_end = std::min(b_beg + max_block_size, sb_end);

    pair_sorter_type &ls = *lcp_sorters[b_id];
    ls.sort(ls.min_merge_budget());

    for (long i = b_beg, lcp; i < b_end; ++i, ++isa, prev = lcp, ++cnt) {
      if (!ls.empty() && ls->first.ll() == i) {  // PLCP[i] is irreducible
        lcp = ls->second;
        ++ls;
        ++irr;
        irr_lcp_sum += lcp;
      } else {  // PLCP[i] is reducible
        lcp = std::max(0L, prev - 1);
      }

      // At this point:
      //   PLCP[i] = lcp
      //   ISA[i] = *isa
      // thus LCP[*isa] = lcp
      lex_sorter->push(Pair(*isa, lcp));
      lcp_sum += lcp;

      if (!(i & kPrintStep)) {
        long double elapsed = utils::wclock() - start;
        long double io_volume = 15L * cnt + 10L * irr;
        long double io_speed = (io_volume / kMiB) / elapsed;
        fprintf(stderr, "\r  Permuting PLCP into LCP: %.1Lf%%, time = %.1Lfs, "
            "io = %.2LfMiB/s", (100.L * cnt) / sb_size, elapsed, io_speed);
      }
    }

    ls.finish_clear();
    delete lcp_sorters[b_id++];
  }

  stxxl::block_manager *bm = stxxl::block_manager::get_instance();
  bm->delete_blocks(isa_bids->begin(), isa_bids->end());
  delete isa_bids;
  delete[] lcp_sorters;

  long double elapsed = utils::wclock() - start;
  long double io_speed = ((15L * cnt + 10L * irr) / kMiB) / elapsed;
  fprintf(stderr, "\r  Permuting PLCP into LCP: "
        "time = %.1Lfs "
        "(+%.3Lfs/MiB), "
        "io = %.2LfMiB/s, "
        "max-alloc = %.2LfB/B\n",
        elapsed,
        elapsed / (text_length / kMiB),
        io_speed,
        1.L * max_alloc() / text_length);

  fprintf(stderr, "\r  Writing LCP to disk:");
  start = utils::wclock();

  typedef stxxl::typed_block<kDiskBlockSize, uint40> tb;
  long lcp_blocks = (sb_size + tb::size - 1) / tb::size;
  lcp_bids[sb_id] = new stxxl::BIDArray<kDiskBlockSize>(lcp_blocks);

  bm->new_blocks(stxxl::striping(), lcp_bids[sb_id]->begin(),
      lcp_bids[sb_id]->end());
  stxxl::buf_ostream<tb, stxxl::BIDArray<kDiskBlockSize>::iterator>
    lcp_ostream(lcp_bids[sb_id]->begin(), 2);

  // Write PLCP[sb_beg..sb_end) to disk permuted to lex order.
  lex_sorter->sort(ram_use);
  for (long i = 0L; i < sb_size; ++i, ++(*lex_sorter)) {
    //   (*lex_sorter)->first = j for some j s.t. sa_beg <= SA[j] < sb_end
    //   (*lex_sorter)->second = LCP[j]
    lcp_ostream << (*lex_sorter)->second;

    if (!(i & kPrintStep)) {
      long double elapsed = utils::wclock() - start;
      long double io_speed = ((15L * i) / kMiB) / elapsed;
      fprintf(stderr, "\r  Writing LCP to disk: %.1Lf%%, time = %.1Lfs, "
          "io = %.2LfMiB/s", (100.L * i) / sb_size, elapsed, io_speed);
    }
  }
  lcp_ostream.fill(0L);

  lex_sorter->finish_clear();
  delete lex_sorter;

  elapsed = utils::wclock() - start;
  io_speed = ((15L * sb_size) / kMiB) / elapsed;
  fprintf(stderr, "\r  Writing LCP to disk: "
      "time = %.1Lfs "
      "(+%.3Lfs/MiB), "
      "io = %.2LfMiB/s, "
      "max-alloc = %.2LfB/B\n",
      elapsed,
      elapsed / (text_length / kMiB),
      io_speed,
      1.L * max_alloc() / text_length);
}

//==============================================================================
// Merge (with the help of SA) LCP subsequences stored on disk into the final
// LCP array and write the result to disk.
//==============================================================================
void merge(std::string sa_fname, std::string out_fname, long text_length,
    long ram_use, long *sb_sizes,
    stxxl::BIDArray<kDiskBlockSize>** &lcp_bids) {
  fprintf(stderr, "Merging: ");
  long double merge_start = utils::wclock();

  stream_reader<uint40> *sa_reader =
    new stream_reader<uint40>(sa_fname, ram_use / 2);
  stream_writer<uint40> *lcp_writer =
    new stream_writer<uint40>(out_fname, ram_use / 2);

  typedef stxxl::typed_block<kDiskBlockSize, uint40> tb;
  typedef stxxl::buf_istream<tb, stxxl::BIDArray<kDiskBlockSize>::iterator>
    istream_type;

  istream_type *readers[kSBlockCnt];
  for (long i = 0; i < kSBlockCnt; ++i)
    readers[i] = new istream_type(lcp_bids[i]->begin(), lcp_bids[i]->end(), 2);

  long sb_beg[kSBlockCnt];
  sb_beg[0] = 0L;
  for (long i = 0L; i + 1 < kSBlockCnt; ++i)
    sb_beg[i + 1] = sb_beg[i] + sb_sizes[i];

  for (long i = 0L; i < text_length; ++i) {
    long sai = sa_reader->read(), sb_id = 0L;
    while (sb_id + 1 < kSBlockCnt && sb_beg[sb_id + 1] <= sai) ++sb_id;

    uint40 lcp;
    *readers[sb_id] >> lcp;
    lcp_writer->write(lcp);
    max_lcp = std::max(max_lcp, lcp.ll());

    if (!(i & kPrintStep)) {
      long double elapsed = utils::wclock() - merge_start;
      long double io_speed = (15L * i / kMiB) / elapsed;
      fprintf(stderr, "\rMerging: "
          "%.1Lf%%, "
          "time = %.1Lfs, "
          "io = %.2LfMiB/s",
          (100.L * i) / text_length,
          elapsed,
          io_speed);
    }
  }

  delete sa_reader;
  delete lcp_writer;

  stxxl::block_manager *bm = stxxl::block_manager::get_instance();
  for (long i = 0L; i < kSBlockCnt; ++i) {
    bm->delete_blocks(lcp_bids[i]->begin(), lcp_bids[i]->end());
    delete lcp_bids[i];
    delete readers[i];
  }

  delete[] lcp_bids;

  long double merge_time = utils::wclock() - merge_start;
  fprintf(stderr, "\rMerging: 100.0%%, "
      "time = %.1Lfs "
      "(+%.3Lfs/MiB), "
      "io = %.2LfMiB/s, "
      "max-alloc = %.2LfB/B\n",
      merge_time,
      merge_time / (text_length / kMiB),
      (15L * text_length / kMiB) / merge_time,
      1.L * max_alloc() / text_length);
}

//==============================================================================
// Build LCP array in internal memory.
//==============================================================================
void im_laca(std::string text_fname, std::string sa_fname,
    std::string out_fname, long length) {
  long *tab = new long[length];

  // Compute Phi and store in tab.
  stream_reader<uint40> *sa_reader = new stream_reader<uint40>(sa_fname);
  for (long i = 0, prev = length; i < length; ++i) {
    long sa_i = sa_reader->read();
    tab[sa_i] = prev;
    prev = sa_i;
  }
  delete sa_reader;

  // Read text.
  unsigned char *text = new unsigned char[length];
  utils::read_objects_at_offset<unsigned char>(text_fname, 0L, length, text);

  // Compute PLCP (overwrite Phi).
  for (long i = 0, match = 0; i < length; ++i) {
    long j = tab[i]; // j = Phi[i]
    while (i + match < length && j + match < length &&
        text[i + match] == text[j + match]) ++match;
    tab[i] = match;
    match = std::max(match - 1, 0L);
  }
  delete[] text;

  // Permute PLCP into LCP and write to disk.
  sa_reader = new stream_reader<uint40>(sa_fname);
  stream_writer<uint40> *lcp_writer = new stream_writer<uint40>(out_fname);
  for (long i = 0; i < length; ++i) {
    long sa_i = sa_reader->read();
    long lcp_i = tab[sa_i];
    lcp_writer->write(lcp_i);
  }
  delete sa_reader;
  delete lcp_writer;
  delete[] tab;
}

//==============================================================================
// LCPscan main function.
//==============================================================================
void LCPscan(std::string text_fname, std::string sa_fname,
    std::string out_fname, long length, long ram_use) {
  // Compute superblock sizes.
  long sb_sizes[kSBlockCnt];
  for (long i = 0; i + 1 < kSBlockCnt; ++i)
    sb_sizes[i] = length * kSBlockFractions[i];
  sb_sizes[kSBlockCnt - 1] = length;
  for (long i = 0; i + 1 < kSBlockCnt; ++i)
  sb_sizes[kSBlockCnt - 1] -= sb_sizes[i];

  fprintf(stderr, "\nNumber of superblocks = %ld\n", kSBlockCnt);
  fprintf(stderr, "Superblock sizes:\n");
  for (long i = 0; i < kSBlockCnt; ++i)
    fprintf(stderr, "  #%ld: %ld (%.2LfMiB, %.1Lf%%)\n", i + 1, sb_sizes[i],
        sb_sizes[i] / (1024.L * 1024), (100.L * sb_sizes[i]) / length);
  fprintf(stderr, "\n");

  // Collect critical positions.
  std::list<long> crit_pos;
  compute_crit_pos(text_fname, sa_fname, crit_pos);

  // Process superblocks.
  stxxl::BIDArray<kDiskBlockSize> **lcp_bids =
    new stxxl::BIDArray<kDiskBlockSize>*[kSBlockCnt];
  for (long sb_beg = 0L, id = 0L; id < kSBlockCnt; sb_beg += sb_sizes[id++]) {
    long sb_end = sb_beg + sb_sizes[id], sb_size = sb_sizes[id];

    // Start the timer.
    fprintf(stderr, "Processing superblock %ld/%ld [%ld..%ld)\n", id + 1,
        kSBlockCnt, sb_beg, sb_end);
    long double sb_start = utils::wclock();

    // Sort triples (SA[i], SA[i - 1], i) by the first component. Only
    // triples with SA[i] inside the superblock are considered.
    CmpTriple cmp_triple;
    triple_sorter_type *triple_sorter = new triple_sorter_type(cmp_triple,
        (30L * ram_use) / 46L, 0L);
    fprintf(stderr, "  Budget for sorting triples = %.2LfMiB\n",
        ram_use / kMiB);
 
    long triple_merge_budget = sort_triples(sa_fname, length, sb_beg, sb_end,
      *triple_sorter);
    fprintf(stderr, "  Budget for merging triples = %.2LfMiB\n",
        triple_merge_budget / kMiB);

    long ram_left = ram_use - triple_merge_budget;
    if (ram_left < kDiskBlockSize) {
      fprintf(stderr, "Error: not enough RAM. Increase the memory "
          "limit (option -m) and try again.\n");
      std::exit(EXIT_FAILURE);
    }

    // Compute the maximal block size so that during process_block
    // we can afford to:
    //   (1) keep the block (max_block_size bytes) in memory,
    //   (2) perform one-pass merging of any of pair_sorters returned
    //       by sort_pairs (assuming these sorters were presorted
    //       using ram_left bytes),
    //   (3) have about 60MiB left to sort the pairs resulting from
    //       process_block computation.
    long double mbs = (ram_use - 60L * kDiskBlockSize);  // This avoids
    mbs /= (ram_left + 30.L * kDiskBlockSize);           // overflow of
    long max_block_size = (long)ceill(mbs * ram_left);   // integer types.
    if (max_block_size < kDiskBlockSize) {
      fprintf(stderr, "Error: not enough RAM. Increase the memory "
          "limit (option -m) and try again.\n");
      std::exit(EXIT_FAILURE);
    }

    long n_b = (sb_size + max_block_size - 1) / max_block_size;

    fprintf(stderr, "  Budget for sorting pairs = %.2LfMiB\n", ram_left / kMiB);
    fprintf(stderr, "  Max block size = %.2LfMiB\n", max_block_size / kMiB);

    // Sort pairs (i, Phi[i]) by second component.
    pair_sorter_type **pair_sorters = new pair_sorter_type*[n_b];
    stxxl::BIDArray<kDiskBlockSize> *isa_bids;
    sort_pairs(sb_beg, sb_end, max_block_size, length, triple_sorter,
        pair_sorters, isa_bids, (20L * ram_left) / 31L, crit_pos);

    // Compute irreducible PLCP values inside the superblock.
    fprintf(stderr, "  Computing PLCP:\n");
    pair_sorter_type **lcp_sorters = new pair_sorter_type*[n_b];
    for (long s = sb_beg, t = 0; s < sb_end; s += max_block_size, ++t) {
      long e = std::min(s + max_block_size, sb_end);
      lcp_sorters[t] = process_block(s, e, text_fname, length, pair_sorters[t]);
    }
    delete[] pair_sorters;

    // Add reducible PLCPs and permute (with the help of ISA) to lex-order.
    plcp_to_lcp(sb_beg, sb_end, id, max_block_size, lcp_sorters, isa_bids,
        lcp_bids, ram_use, length);

    // Print superblock summary.
    long double sb_tot_time = utils::wclock() - sb_start;
    fprintf(stderr, "Summary for superblock: "
        "time = %.1Lfs "
        "(+%.3Lfs/MiB), "
        "max-alloc = %.2LfB/B\n\n",
        sb_tot_time,
        sb_tot_time / (length / kMiB),
        1.L * max_alloc() / length);
  }

  // Merge LCP subsequences into final LCP array.
  merge(sa_fname, out_fname, length, ram_use, sb_sizes, lcp_bids);
}

//==============================================================================
// Main dispatching function. Compute the LCP array of text_fname and write
// into out_fname using at most ram_use bytes of RAM.
//==============================================================================
void construct_lcp(std::string text_fname, std::string sa_fname,
    std::string out_fname, long ram_use) {
  long double start = utils::wclock();

  lcp_sum = max_lcp = irr_lcp_sum = n_irr = 0L;
  long length = utils::file_size(text_fname);

  // Force STXXL to print info about disks.
  (void)max_alloc();

  // Turn paths abs lute.
  text_fname = utils::absolute_path(text_fname);
  sa_fname = utils::absolute_path(sa_fname);
  out_fname = utils::absolute_path(out_fname);

  // Print input parameters.
  fprintf(stderr, "\nText filename = %s\n", text_fname.c_str());
  fprintf(stderr, "SA filename = %s\n", sa_fname.c_str());
  fprintf(stderr, "Output filename = %s\n", out_fname.c_str());
  fprintf(stderr, "Text length = %ld (%.2LfMiB)\n", length, length / kMiB);
  fprintf(stderr, "RAM use = %ld (%.2LfMiB)\n", ram_use,
      ram_use / (1024.L * 1024));

  if (length > kSmallInstance) {
    fprintf(stderr, "Running LCPscan...\n");
    LCPscan(text_fname, sa_fname, out_fname, length, ram_use);
  } else {  // for small instances run in-memory algorithm
    fprintf(stderr, "Running internal memory algorithm...\n");
    im_laca(text_fname, sa_fname, out_fname, length);
  }

  // Print final summary and stats.
  long double elapsed = utils::wclock() - start;
  fprintf(stderr, "Total time: %.2Lfs (%.3Lfs/MiB)\n",
      elapsed, elapsed / (length / kMiB));

  if (length > kSmallInstance) {
    fprintf(stderr, "\nStats:\n");
    fprintf(stderr, "  maximal LCP = %ld\n", max_lcp);
    fprintf(stderr, "  sum of LCPs = %ld\n", lcp_sum);
    fprintf(stderr, "  number of irreducible LCPs = %ld\n", n_irr);
    fprintf(stderr, "  sum of irreducible LCPs = %ld\n", irr_lcp_sum);
  }
}
