#include <cstdio>
#include <cstdlib>
#include <getopt.h>

#include <string>

#include "LCPscan/LCPscan.h"
#include "LCPscan/utils.h"

char *program_name;

void usage(int status) {
  printf(
"Usage: %s [OPTION]... FILE\n"
"Construct the LCP array for text stored in FILE.\n"
"\n"
"Mandatory arguments to long options are mandatory for short options too.\n"
"  -m, --mem=LIMIT         limit RAM usage to LIMIT MiB (default: 3072)\n"
"  -o, --output=OUTFILE    specify output file (default: FILE.lcp5)\n"
"  -s, --sa=SUFARRAY       specify the location of the suffix array of FILE\n"
"                          (default: FILE.sa5)\n", program_name);

  std::exit(status);
}

int main(int argc, char **argv) {
  program_name = argv[0];

  static struct option long_options[] = {
    {"mem",    required_argument, NULL, 'm'},
    {"output", required_argument, NULL, 'o'},
    {"sa",     required_argument, NULL, 's'},
    {NULL, 0, NULL, 0}
  };

  long ram_use = 3072L << 20;
  std::string sa_fname("");
  std::string out_fname("");

  // Parse command-line options.
  int c;
  while ((c = getopt_long(argc, argv, "m:o:s:", long_options, NULL)) != -1) {
    switch(c) {
      case 'm':
        ram_use = std::atol(optarg) << 20;
        if (ram_use <= 0L) {
          fprintf(stderr, "Error: invalid RAM limit (%ld)\n\n", ram_use);
          usage(EXIT_FAILURE);
        }
        break;
      case 'o':
        out_fname = std::string(optarg);
        break;
      case 's':
        sa_fname = std::string(optarg);
        break;
      default:
        usage(EXIT_FAILURE);
    }
  }

  if (optind >= argc) {
    fprintf(stderr, "Error: FILE not provided\n\n");
    usage(EXIT_FAILURE);
  }

  // Parse the text filename.
  std::string text_fname = std::string(argv[optind++]);
  if (optind < argc) {
    fprintf(stderr, "Warning: multiple input files provided. "
        "Only the first will be processed.\n\n");
  }

  // Set default SA and output filenames (if not provided).
  if (sa_fname.empty()) sa_fname = text_fname + ".sa5";
  if (out_fname.empty()) out_fname = text_fname + ".lcp5";

  // Check for their existence.
  if (!utils::file_exists(text_fname)) {
    fprintf(stderr, "Error: input file (%s) does not exist\n\n",
        text_fname.c_str());
    usage(EXIT_FAILURE);
  }
  if (!utils::file_exists(sa_fname)) {
    fprintf(stderr, "Error: suffix array (%s) does not exist\n\n",
        sa_fname.c_str());
    usage(EXIT_FAILURE);
  }

  utils::find_stxxl_config();

  construct_lcp(text_fname, sa_fname, out_fname, ram_use);
}
